import logging
import os
import sys
import importlib


def get_version() -> str:
    try:
        return importlib.metadata.version(__package__)
    except importlib.metadata.PackageNotFoundError:
        return 'development'


def load_envs() -> list:
    envs = []
    required_envs = ["ZAMMAD_BASE_URL", "ZAMMAD_TOKEN", "BASIC_AUTH_USER", "BASIC_AUTH_PASSWORD"]
    for required_env in required_envs:
        if required_env not in os.environ:
            logging.fatal(f"Required environment variable not set: {required_env}")
            logging.fatal("Here you will find some docs: https://github.com/kmille/zammad-pgp-auto-import")
            sys.exit(1)
        else:
            envs.append(os.environ[required_env])
    LISTEN_HOST = os.environ.get("LISTEN_HOST", "127.0.0.1")
    LISTEN_PORT = int(os.environ.get("LISTEN_PORT", "22000"))
    DEBUG = os.environ.get("DEBUG", "0")
    return [*envs, LISTEN_HOST, LISTEN_PORT, DEBUG]



import logging
import json
from typing import Optional
from flask import Flask, request
from flask_basicauth import BasicAuth
import waitress
import werkzeug

from zammad_pgp_import.utils import get_version, load_envs
from zammad_pgp_import.exceptions import PGPImportError, NotFoundOnKeyserverError, ZammadPGPKeyAlreadyImportedError
from zammad_pgp_import.zammad import Zammad
from zammad_pgp_import.pgp import PGPHandler, PGPKey

LOG_FORMAT = "[%(asctime)s %(levelname)5s] %(message)s"
logging.basicConfig(format=LOG_FORMAT,
                    level=logging.INFO)

logging.getLogger("urllib3").setLevel(logging.WARNING)


logger = logging.getLogger(__name__)

(ZAMMAD_BASE_URL, ZAMMAD_TOKEN, BASIC_AUTH_USER, BASIC_AUTH_PASSWORD,
    LISTEN_HOST, LISTEN_PORT, DEBUG) = load_envs()

if DEBUG == "1":
    logger.setLevel(logging.DEBUG)

app = Flask(__name__)
app.config['BASIC_AUTH_USERNAME'] = BASIC_AUTH_USER
app.config['BASIC_AUTH_PASSWORD'] = BASIC_AUTH_PASSWORD
app.config['BASIC_AUTH_FORCE'] = True  # protect all endpoints
basic_auth = BasicAuth(app)

error_counter = 0


def is_encrypted_mail(article_data: dict) -> bool:
    preferences = article_data["preferences"]
    if "security" not in preferences:
        return False

    security = preferences["security"]
    if security.get("type", "") != "PGP":
        logger.debug("Email seems encrypted, but not with PGP")
        return False

    try:
        return security["encryption"].get("success", False) is True
    except KeyError as e:
        logger.error(f"Could not check if mail is PGP encrypted: {e}")
        logger.error(json.dumps(article_data, indent=4))
        return False


def get_pgp_key_from_attachments(article_data: dict) -> Optional[PGPKey]:
    # this only supports a single PGP key
    if len(article_data["attachments"]) == 0:
        logger.debug("This ticket does not have any attachments")
        return None
    for attachment in article_data["attachments"]:
        if "application/pgp-keys" in attachment["preferences"].get("Content-Type", ""):
            logger.debug("Seems like a PGP key is attached to this email")
            z = Zammad(ZAMMAD_BASE_URL, ZAMMAD_TOKEN)
            key_data = z.download_attachment(attachment["url"])
            return PGPKey(key_data)
    return None


def import_pgp_key(pgp_key: PGPKey, sender_email: str) -> None:
    if not pgp_key.has_email(sender_email):
        logger.warning(f"E-Mail contains a PGP not matching with senders email ({sender_email}, {pgp_key})")
    elif pgp_key.is_expired:
        pass
        # logger.warning(f"PGP key is already expired. Not importing it ({pgp_key.expires})")
    else:
        z = Zammad(ZAMMAD_BASE_URL, ZAMMAD_TOKEN)
        try:
            z.import_pgp_key(pgp_key)
        except ZammadPGPKeyAlreadyImportedError as e:
            logger.info(e)
        else:
            logger.info(f"Successfully imported pgp key ({pgp_key.fingerprint}) for email {sender_email}")


def get_key_from_keyserver(email: str) -> Optional[PGPKey]:
    try:
        pgp_key = PGPHandler.search_pgp_key(email)
        logger.info("Successfully found PGP key using a keyserver")
        return pgp_key
    except NotFoundOnKeyserverError as e:
        logging.error(e)
        return None


@app.route("/api/zammad/pgp", methods=["POST"])
def webhook_new_ticket() -> tuple[dict[str, str], int]:
    global error_counter
    try:
        data = request.get_json(force=True)
        sender_email = data["ticket"]["created_by"]["email"]
        article_data = data['article']

        is_encrypted = is_encrypted_mail(article_data)
        logger.info(f"Received a new Ticket: {ZAMMAD_BASE_URL}/#ticket/zoom/{data['ticket']['id']} (from={sender_email}, is_encrypted={is_encrypted})")

        pgp_key = get_pgp_key_from_attachments(article_data)
        if not pgp_key and is_encrypted:
            pgp_key = get_key_from_keyserver(sender_email)
        if pgp_key:
            import_pgp_key(pgp_key, sender_email)
        else:
            logger.info("Ticket does not have a PGP key attached. It is also not encrypted and/or PGP key was not found on a keysever")

    except (werkzeug.exceptions.BadRequest, KeyError) as e:
        logger.exception(e)
        error_counter += 1
        return {"status": 'failed', 'reason': 'invalid client request'}, 500
    except PGPImportError as e:
        logger.error(e)
        error_counter += 1
        return {"status": 'failed', 'reason': 'PGP/API error'}, 500
    except Exception as e:
        logger.error(f"Unhandled exception occured: {e}")
        logger.exception(e)
        error_counter += 1
        return {"status": "failed", 'reason': 'unhandled exception'}, 500
    else:
        return {"status": "ok"}, 200


@app.route("/status")
def status() -> dict:
    if error_counter == 0:
        return {"status": "ok"}
    else:
        return {"status": "failed"}


def main() -> None:
    logger.info(f"Starting webhook backend on {LISTEN_HOST}:{LISTEN_PORT} (version {get_version()}, debug={DEBUG == '1'})")
    if __name__ == '__main__':
        app.run(host=LISTEN_HOST, port=LISTEN_PORT, debug=True)
    else:
        waitress.serve(app, listen=f"{LISTEN_HOST}:{LISTEN_PORT}")


if __name__ == '__main__':
    main()



import logging
import re
import os
import subprocess
# from datetime import datetime, date
import requests

from zammad_pgp_import.exceptions import PGPError, NotFoundOnKeyserverError

logger = logging.getLogger(__name__)

KEY_SERVER = os.environ.get("KEY_SERVER", "https://keys.openpgp.org")


class PGPKey(object):
    raw: str
    meta: str
    emails: list[str] = []
    fingerprint: str
    # expires: date
    is_expired: bool

    def __init__(self, key_data: str):
        self.raw = key_data
        try:
            p = subprocess.run(["gpg", "--show-keys"], input=key_data.encode(), capture_output=True, check=True)
            self.meta = p.stdout.decode()
        except FileNotFoundError as e:
            raise PGPError(f"Could not find gpg binary: {e}")
        except subprocess.CalledProcessError as e:
            logging.error(f"Content of the PGP key\n: {key_data}")
            raise PGPError(f"Could not decode pgp key: {e.stderr.decode()}")

        for line in self.meta.splitlines():
            if line.startswith("uid"):
                if result := re.search(r'<(.*)>', line):
                    self.emails.append(result.group(1))
            elif result := re.search(r'[0-9A-F]{40}', line):
                self.fingerprint = result[0]
            # elif result := re.search(r'expires: (\d{4}-\d{2}-\d{2})', line):
            #     self.expires = datetime.strptime(result.group(1), "%Y-%m-%d").date()
            #     self.is_expired = datetime.now().date() > self.expires

        #  TODO: test if everything is set up
        # breakpoint()
        # if not all([self.fingerprint, self.expires]):
        #     raise PGPError("Could not parse PGP key data")
        self.is_expired = False

    def has_email(self, email: str) -> bool:
        return email in self.emails

    def __repr__(self) -> str:
        # return f"PGPKey (emails={','.join(self.emails)} fingerprint={self.fingerprint}, expires={self.expires}))"
        return f"PGPKey (emails={','.join(self.emails)} fingerprint={self.fingerprint}"


class PGPHandler:

    @staticmethod
    def search_pgp_key(email: str) -> PGPKey:
        logging.debug(f"Using keyserver {KEY_SERVER} to find a PGP key for {email}")
        try:
            resp = requests.get(KEY_SERVER + f"/pks/lookup?op=get&options=mr&search={email}")
            resp.raise_for_status()
            return PGPKey(resp.text)
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 400:
                logger.debug(f"API error message: {e.response.text}")
                raise NotFoundOnKeyserverError(f"Could not find a PGP key for {email} using keyserver {KEY_SERVER}")
            else:
                raise PGPError(f"Could find PGP key on {KEY_SERVER}: {e.response.text}")
        except requests.exceptions.RequestException as e:
            raise PGPError(f"Could not get PGP key: {e}")



import logging
import json
import requests

from zammad_pgp_import.utils import get_version
from zammad_pgp_import.pgp import PGPKey
from zammad_pgp_import.exceptions import ZammadError, ZammadPGPKeyAlreadyImportedError

logger = logging.getLogger(__name__)


class Zammad(object):
    session: requests.Session
    bsae_url: str

    def __init__(self, zammad_base_url: str, auth_token: str):
        self.base_url = zammad_base_url
        self.session = requests.Session()
        self.session.headers.update({"Authorization": f"Token token={auth_token}",
                                     "User-Agent": f"PGP auto-import webhook {get_version()}"})

    def get_all_imported_pgp_keys(self) -> list:
        """
        This is not used any more.  We just import and do net check if the key already exists before
        """
        logger.debug("Getting all imported PGP keys using Zammad API")
        try:
            req = self.session.get(self.base_url + "/api/v1/integration/pgp/key")
            req.raise_for_status()
            return req.json()
        except requests.exceptions.RequestException as e:
            raise ZammadError(f"Could not get all imported PGP keys from Zammad: {e}")

    def download_attachment(self, url: str) -> str:
        logger.debug("Downloading ticket attachment using Zammad API")
        try:
            resp = self.session.get(url)
            resp.raise_for_status()
            logger.debug("Successfully downloaded email attachment")
            return resp.text
        except requests.exceptions.RequestException as e:
            if isinstance(e, requests.exceptions.HTTPError):
                logger.error(f"Zammad API error: {e}")
                raise ZammadError(f"Could not download attachment from Zammad: {e.response.json()['error_human']}")
            raise ZammadError(f"Could not download attachment from Zammad: {e}")

    def import_pgp_key(self, pgp_key: PGPKey) -> None:
        logger.debug("Importing PGP key using Zammad API")
        data = {'file': '',
                'key': pgp_key.raw,
                'passphrase': ""}
        try:
            resp = self.session.post(self.base_url + "/api/v1/integration/pgp/key", json=data)
            resp.raise_for_status()
        except requests.exceptions.RequestException as e:
            if isinstance(e, requests.exceptions.HTTPError) and e.response.status_code == 422:
                raise ZammadPGPKeyAlreadyImportedError(f"Key was already imported. API response: {e.response.json()['error_human']}")
            elif isinstance(e, requests.exceptions.HTTPError):
                logger.error(f"Zammad API error: {e}")
                logger.error(f"Request json:\n{json.dumps(data, indent=4)}")
                raise ZammadError(f"Could not import PGP key: {e.response.json()['error_human']}")
            raise ZammadError(f"Could not import PGP key: {e}")

    def delete_pgp_key(self, email: str) -> None:
        logger.debug("Deleting PGP key using Zammad API")
        try:
            all_imported_keys = self.get_all_imported_pgp_keys()
            matching_keys = list(filter(lambda x: x['email_addresses'][0].lower() == email.lower(), all_imported_keys))
            if len(matching_keys) == 0:
                logger.warning(f"Could find a PGP key with e-mail {email}")
            else:
                resp = self.session.delete(self.base_url + f"/api/v1/integration/pgp/key/{matching_keys[0]['id']}")
                resp.raise_for_status()
                logger.info("Successfully deleted PGP in Zammad")
        except requests.exceptions.RequestException as e:
            logger.warning(f"Could not delete PGP key in Zammad: {e}")




class PGPImportError(Exception):
    pass


class PGPError(PGPImportError):
    pass


class NotFoundOnKeyserverError(PGPImportError):
    pass


class ZammadError(PGPImportError):
    pass


class ZammadPGPKeyAlreadyImportedError(PGPImportError):
    pass
